;*******************************************************************************
;*
;*	Titel:	Unterprogrammbibliothek zur Ganzzahlmultiplikation und
;*      Division (SR_MUL_DIV.asm)
;*	Datum:	15/05/10	Version:	0.3
;*	Autor:	WEIGU
;*
;*	Informationen zur Beschaltung:
;*      
;*	Prozessor: ATmega32			Quarzfrequenz:	
;*	Eingaenge:
;*	Ausgaenge:
;*
;*	Informationen:
;* 
;*      DIV21   Ganzahldivision 2Byte/1Byte   
;*      DIV32   Ganzahldivision 3Byte/2Byte   
;*      DIV42   Ganzahldivision 4Byte/2Byte
;*      DIV62   Ganzahldivision 6Byte/2Byte
;*      MUL22   Ganzahlmultiplikation 2Byte*2Byte   
;*      MUL42   Ganzahlmultiplikation 2Byte*2Byte   
;*      
;*      Copyright (c) 2010   Guy WEILER         weigu[at]weigu[dot]lu
;*
;*      Hiermit wird unentgeltlich, jeder Person, die eine Kopie der Software
;*      und der zugehoerigen Dokumentationen (die "Software") erhaelt, die
;*      Erlaubnis erteilt, uneingeschraenkt zu benutzen, inklusive und ohne
;*      Ausnahme, dem Recht, sie zu verwenden, kopieren, aendern, fusionieren,
;*      verlegen, verbreiten, unterlizenzieren und/oder zu verkaufen, und
;*      Personen, die diese Software erhalten, diese Rechte zu geben, unter den
;*      folgenden Bedingungen:
;*
;*      ;Der obige Urheberrechtsvermerk und dieser Erlaubnisvermerk sind in alle
;*      ;Kopien oder Teilkopien der Software beizulegen.
;*      ;
;*      ;DIE SOFTWARE WIRD OHNE JEDE AUSDRUECKLICHE ODER IMPLIZIERTE GARANTIE
;*      ;BEREITGESTELLT, EINSCHLIESSLICH DER GARANTIE ZUR BENUTZUNG FUER DEN
;*      ;VORGESEHENEN ODER EINEM BESTIMMTEN ZWECK SOWIE JEGLICHER RECHTS-
;*      ;VERLETZUNG, JEDOCH NICHT DARAUF BESCHRAENKT. IN KEINEM FALL SIND DIE
;*      ;AUTOREN ODER COPYRIGHTINHABER FUER JEGLICHEN SCHADEN ODER SONSTIGE
;*      ;ANSPRUECHE HAFTBAR ZU MACHEN, OB INFOLGE DER ERFUELLUNG EINES
;*      ;VERTRAGES, EINES DELIKTES ODER ANDERS IM ZUSAMMENHANG MIT DER SOFTWARE
;*      ;ODER SONSTIGER VERWENDUNG DER SOFTWARE ENTSTANDEN.
;*			
;*******************************************************************************
;Damit der Code im Flash gering bleibt werden ausschliesslich die benutzten
;Unterprogramme freigeschaltet (Flag = 1)
.EQU    Flag_DIV21  = 1		
.EQU    Flag_DIV32  = 1		
.EQU    Flag_DIV42  = 1		
.EQU    Flag_DIV62  = 0		
.EQU    Flag_MUL22  = 1		
.EQU    Flag_MUL42  = 1		
.EQU    Flag_MUL22S = 1		
.EQU    Flag_MUL42S = 1		

;-------------------------------------------------------------------------------
;       Ganzzahldivision 2Byte / 1Byte
;-------------------------------------------------------------------------------

;Dividend:       r1,r0 (LSB)
;Divisor         r2
;Schleifenzähler r16
;Resultat Rest in r0 + Quotient in r1
;Achtung Ueberlauf wenn obere n Bit des Dividenden groesser gleich n Bit Divisor

.IF Flag_DIV21 == 1
DIV21:  push    r16
        in      r16,SREG
        push    r16
        
        ldi     r16,8           ;Schleifenzaehler init. (Bitzahl Quotient)
        
DIV21A: lsl     r0              ;Rotiere Dividend 1 Stelle nach links
        rol     r1              ;rechts wird Null reingeschoben
        brcs    DIV21B          ;Falls Carry muss Subtraktion erfolgen!
        cp      r1,r2           ;Subtraktion testen
        brlo    DIV21C          ;Ueberspringe Subtraktion, wenn kleiner

DIV21B: sub     r1,r2           ;Subtraktion durchfuehren
        inc     r0              ;Bit 0 im Dividenden setzen

DIV21C: dec     r16             ;Schleifenzaehler dekrementieren
        brne    DIV21A

        pop     r16
        out     SREG,r16
        pop     r16
        ret
.ENDIF

;-------------------------------------------------------------------------------
;       Ganzzahldivision 3Byte / 2Byte
;-------------------------------------------------------------------------------
;Dividend:       r2,r1,r0 (LSB)
;Divisor         r4,r3 (LSB)
;Schleifenzähler r16
;Resultat: Rest (r2:r1) + Quotient (r0) in Dividendenregister
;Achtung Ueberlauf wenn obere n Bit des Dividenden groesser gleich n Bit Divisor

.IF Flag_DIV32 == 1
DIV32:  push    r16
        in      r16,SREG
        push    r16

        ldi     r16,8           ;Schleifenzaehler SZ init. (Bitzahl Quotient)

DIV32A: lsl     r0              ;Rotiere Dividend 1 Stelle nach links
        rol     r1              ;rechts wird Null reingeschoben
        rol     r2
        brcs    DIV32B          ;Falls Carry muss Subtraktion erfolgen!
        cp      r1,r3           ;Subtraktion testen
        cpc     r2,r4
        brlo    DIV32C          ;Ueberspringe Subtraktion, wenn kleiner

DIV32B: sub     r1,r3           ;Subtraktion durchführen
        sbc     r2,r4
        inc     r0              ;Bit 0 im Dividenden setzen

DIV32C: dec     r16             ;Schleifenzaehler dekrementieren
        brne    DIV32A          ;Bleib in der Schleife bis SZ = 0

        pop     r16             ;Register wiederherstellen
        out     SREG,r16
        pop     r16
        ret                     ;Zurueck
.ENDIF

;-------------------------------------------------------------------------------
;       Ganzzahldivision 4Byte / 2Byte
;-------------------------------------------------------------------------------
;Dividend:       r3,r2,r1,r0 (LSB)
;Divisor         r5,r4 (LSB)
;Schleifenzähler r16
;Resultat: Rest (r3:r2) + Quotient (r1:r0) in Dividendenregister
;Achtung Ueberlauf wenn obere n Bit des Dividenden groesser gleich n Bit Divisor

.IF Flag_DIV42 == 1
DIV42:  push    r16
        in      r16,SREG
        push    r16
        
        ldi     r16,16           ;Schleifenzaehler SZ init. (Bitzahl Quotient)
        
DIV42A: lsl     r0              ;Rotiere Dividend 1 Stelle nach links
        rol     r1              ;rechts wird Null reingeschoben
        rol     r2
        rol     r3
        brcs    DIV42B          ;Falls Carry muss Subtraktion erfolgen!
        cp      r2,r4           ;Subtraktion testen
        cpc     r3,r5
        brlo    DIV42C          ;Ueberspringe Subtraktion, wenn kleiner

DIV42B: sub     r2,r4           ;Subtraktion durchführen
        sbc     r3,r5
        inc     r0              ;Bit 0 im Dividenden setzen

DIV42C: dec     r16             ;Schleifenzaehler dekrementieren
        brne    DIV42A          ;Bleib in der Schleife bis SZ = 0

        pop     r16             ;Register wiederherstellen
        out     SREG,r16
        pop     r16
        ret                     ;Zurueck
.ENDIF

;-------------------------------------------------------------------------------
;       Ganzzahldivision 6Byte / 2Byte
;-------------------------------------------------------------------------------
;Dividend:       r5,r4,r3,r2,r1,r0 (LSB)
;Divisor         r7,r6 (LSB)
;Schleifenzähler r16
;Resultat: Rest (r5:r4) + Quotient (r3:r0) in Dividendenregister
;Achtung Ueberlauf wenn obere n Bit des Dividenden groesser gleich n Bit Divisor

.IF Flag_DIV62 == 1
DIV62:  push    r16
        in      r16,SREG
        push    r16
        
        ldi     r16,32          ;Schleifenzaehler SZ init. (Bitzahl Quotient)
        
DIV62A: lsl     r0              ;Rotiere Dividend 1 Stelle nach links
        rol     r1              ;rechts wird Null reingeschoben
        rol     r2
        rol     r3
        rol     r4
        rol     r5
        brcs    DIV62B          ;Falls Carry muss Subtraktion erfolgen!
        cp      r4,r6           ;Subtraktion testen
        cpc     r5,r7
        
        brlo    DIV62C          ;Ueberspringe Subtraktion, wenn kleiner

DIV62B: sub     r4,r6           ;Subtraktion durchführen
        sbc     r5,r7
        inc     r0              ;Bit 0 im Dividenden setzen

DIV62C: dec     r16             ;Schleifenzaehler dekrementieren
        brne    DIV62A          ;Bleib in der Schleife bis SZ = 0

        pop     r16             ;Register wiederherstellen
        out     SREG,r16
        pop     r16
        ret                     ;Zurueck
.ENDIF

;-------------------------------------------------------------------------------
;       Ganzzahlmultiplikation 2Byte * 2Byte
;-------------------------------------------------------------------------------
;Multiplikationsbefehl des ATmega benoetigt r0 und r1 fuer Resultat
;Resultat der Multiplikation in r5,r4,r3,r2 (LSB)
;1. Zahl (a)in r7,r6 (LSB)
;2. Zahl (b) in  r9,r8 (LSB)

.IF Flag_MUL22 == 1
MUL22:  push    r16
        in      r16,SREG
        push    r16

        mul     r6,r8           ;Ergebnisregister mit 2 Produkten r6r8 und r7r9 
        movw    r2,r0           ;fuellen 
        mul     r7,r9 
        movw    r4,r0           
        mul     r6,r9           ;Produkt r6r7 versetzt zum Ergebnis addieren
        add     r3,r0
        adc     r4,r1
        adc     r5,Zero         ;Carry beruecksichtigen
        mul     r7,r8           ;Produkt r6r7 versetzt zum Ergebnis addieren
        add     r3,r0
        adc     r4,r1
        adc     r5,Zero

        pop     r16
        out     SREG,r16        ;Statusregister wiederherstellen
        pop     r16
        ret                     ;Zurueck
.ENDIF        

;-------------------------------------------------------------------------------
;	Ganzzahlmultiplikation 4Byte * 2Byte
;-------------------------------------------------------------------------------

;Multiplikationsbefehl des ATmega benoetigt r0 und r1 fuer Resultat
;Resultat der Multiplikation in r7,r6,r5,r4,r3,r2 (LSB)
;1. Zahl (a) in r11,r10,r9,r8 (LSB)
;2. Zahl (b) in  r13,r12 (LSB)

.IF Flag_MUL42 == 1
MUL42:  push    r16
        in      r16,SREG
        push    r16        
        
        mul     r8,r12          ;Ergebnisregister mit 2 Produkten r8r12 und 
        movw    r2,r0           ;r11r13 und Null fuellen
        mul     r11,r13 
        movw    r6,r0           ;movw    r7:r6,r1:r0     
        clr     r4
        clr     r5           
        
        mul     r8,r13          ;Produkt r8r13 versetzt zum Ergebnis addieren
        add     r3,r0
        adc     r4,r1
        adc     r5,Zero         ;Carry beruecksichtigen
        mul     r9,r12          ;Produkt r9r12 versetzt zum Ergebnis addieren
        add     r3,r0
        adc     r4,r1
        adc     r5,Zero         ;Carry beruecksichtigen
        
        mul     r10,r12         ;Produkt r10r12 versetzt zum Ergebnis addieren
        add     r4,r0
        adc     r5,r1
        adc     r6,Zero         ;Carry beruecksichtigen
        mul     r9,r13          ;Produkt r9r13 versetzt zum Ergebnis addieren
        add     r4,r0
        adc     r5,r1
        adc     r6,Zero         ;Carry beruecksichtigen
        
        mul     r11,r12         ;Produkt r11r12 versetzt zum Ergebnis addieren
        add     r5,r0
        adc     r6,r1
        adc     r7,Zero         ;Carry beruecksichtigen
        mul     r10,r13         ;Produkt r10r13 versetzt zum Ergebnis addieren
        add     r5,r0
        adc     r6,r1
        adc     r7,Zero         ;Carry beruecksichtigen
        
        pop     r16
        out     SREG,r16        ;Statusregister wiederherstellen
        pop     r16
        ret                     ;Zurueck
.ENDIF

;-------------------------------------------------------------------------------
;       Ganzzahlmultiplikation signed 
;       2Byte * 2Byte Achtung max. 32767*32767 (bzw. -32768)
;-------------------------------------------------------------------------------
;Multiplikationsbefehl des ATmega benoetigt r0 und r1 fuer Resultat
;Resultat der Multiplikation in r5,r4,r3,r2 (LSB)
;1. Zahl (a)in r7,r6 (LSB)
;2. Zahl (b) in  r9,r8 (LSB)

.IF Flag_MUL22S == 1
MUL22S: push    r16
        in      r16,SREG
        push    r16
        push    r17

        clt
        ldi	r17,0x80	;Ueberpruefe Bit7 der 1. Zahl (negative Flag)
        and	r17,r7
        brne	MUL221
        rjmp	MUL222
MUL221: set			;T-Flag zeigt, dass Multiplikation negativ wird      
        com 	r6		;Zweierkomplement der 1. Zahl
        com 	r7
        ldi	r16,1
        add	r6,r16
        adc	r7,Zero
        
MUL222: ldi	r17,0x80	;Ueberpruefe Bit7 der 1. Zahl (negative Flag)
        and	r17,r9		;Ueberpruefe Bit7 der 2. Zahl (negative Flag)
        brne	MUL223
      	rjmp	MUL225		;positiv dann weiter         

MUL223: com 	r8		;2 Zahl negativ: Zweierkomplement der 2. Zahl
        com 	r9   
        ldi	r16,1
        add	r8,r16
        adc	r9,Zero
        brtc	MUL224		
 	clt			;Falls T-Flag schon gesetzt, dann loeschen
 	rjmp	MUL225		 
MUL224: set

MUL225:	mul     r6,r8           ;Ergebnisregister mit 2 Produkten r6r8 und r7r9 
        movw    r2,r0           ;fuellen 
        mul     r7,r9 
        movw    r4,r0           
        mul     r6,r9           ;Produkt r6r7 versetzt zum Ergebnis addieren
        add     r3,r0
        adc     r4,r1
        adc     r5,Zero         ;Carry beruecksichtigen
        mul     r7,r8           ;Produkt r6r7 versetzt zum Ergebnis addieren
        add     r3,r0
        adc     r4,r1
        adc     r5,Zero
        
        brtc	MUL226		
        com 	r2		;Resultat negativ! Zweierkomplement
        com 	r3
        com 	r4
        com 	r5
        ldi	r16,1
        add	r2,r16
        adc	r3,Zero
        adc	r4,Zero
        adc	r5,Zero

MUL226: pop	r17
        pop     r16
        out     SREG,r16        ;Statusregister wiederherstellen
        pop     r16
        ret                     ;Zurueck
.ENDIF        

;-------------------------------------------------------------------------------
;	Ganzzahlmultiplikation  signed
;	4Byte * 2Byte Achtung max. 2147483647*32767 (bzw. -32768, -2147483648)
;-------------------------------------------------------------------------------

;Multiplikationsbefehl des ATmega benoetigt r0 und r1 fuer Resultat
;Resultat der Multiplikation in r7,r6,r5,r4,r3,r2 (LSB)
;1. Zahl (a) in r11,r10,r9,r8 (LSB)
;2. Zahl (b) in  r13,r12 (LSB)

.IF Flag_MUL42S == 1
MUL42S: push    r16
        in      r16,SREG
        push    r16        
        push    r17

        clt
        ldi	r17,0x80	;Ueberpruefe Bit7 der 1. Zahl (negative Flag)
        and	r17,r11
        brne	MUL421
        rjmp	MUL422
MUL421: set			;T-Flag zeigt, dass Multiplikation negativ wird      
        com 	r8		;Zweierkomplement der 1. Zahl
        com 	r9
        com 	r10
        com 	r11
        ldi	r16,1
        add	r8,r16
        adc	r9,Zero
        adc	r10,Zero
        adc	r11,Zero
        
MUL422: ldi	r17,0x80	;Ueberpruefe Bit7 der 1. Zahl (negative Flag)
        and	r17,r13		;Ueberpruefe Bit7 der 2. Zahl (negative Flag)
        brne	MUL423
      	rjmp	MUL425		;positiv dann weiter         

MUL423: com 	r12		;2 Zahl negativ: Zweierkomplement der 2. Zahl
        com 	r13   
        ldi	r16,1
        add	r12,r16
        adc	r13,Zero
        brtc	MUL424		
 	clt			;Falls T-Flag schon gesetzt, dann loeschen
 	rjmp	MUL425		 
MUL424: set

MUL425: mul     r8,r12          ;Ergebnisregister mit 2 Produkten r8r12 und 
        movw    r2,r0           ;r11r13 und Null fuellen
        mul     r11,r13 
        movw    r6,r0           ;movw    r7:r6,r1:r0     
        clr     r4
        clr     r5           
        
        mul     r8,r13          ;Produkt r8r13 versetzt zum Ergebnis addieren
        add     r3,r0
        adc     r4,r1
        adc     r5,Zero         ;Carry beruecksichtigen
        mul     r9,r12          ;Produkt r9r12 versetzt zum Ergebnis addieren
        add     r3,r0
        adc     r4,r1
        adc     r5,Zero         ;Carry beruecksichtigen
        
        mul     r10,r12         ;Produkt r10r12 versetzt zum Ergebnis addieren
        add     r4,r0
        adc     r5,r1
        adc     r6,Zero         ;Carry beruecksichtigen
        mul     r9,r13          ;Produkt r9r13 versetzt zum Ergebnis addieren
        add     r4,r0
        adc     r5,r1
        adc     r6,Zero         ;Carry beruecksichtigen
        
        mul     r11,r12         ;Produkt r11r12 versetzt zum Ergebnis addieren
        add     r5,r0
        adc     r6,r1
        adc     r7,Zero         ;Carry beruecksichtigen
        mul     r10,r13         ;Produkt r10r13 versetzt zum Ergebnis addieren
        add     r5,r0
        adc     r6,r1
        adc     r7,Zero         ;Carry beruecksichtigen
        
        brtc	MUL426		
        com 	r2		;Resultat negativ! Zweierkomplement
        com 	r3
        com 	r4
        com 	r5
        com 	r6
        com 	r7
        ldi	r16,1
        add	r2,r16
        adc	r3,Zero
        adc	r4,Zero
        adc	r5,Zero
        adc	r6,Zero
        adc	r7,Zero

MUL426: pop	r17        
        pop     r16
        out     SREG,r16        ;Statusregister wiederherstellen
        pop     r16
        ret                     ;Zurueck
.ENDIF