;*******************************************************************************
;*
;*	Titel:	usbammeter (usbammeter.asm)
;*	Datum:	06/05/12	Version:	0.1	
;*	Autor:	WEIGU
;*
;*	Informationen zur Beschaltung:
;*      
;*	Prozessor: ATmega32U4		Quarzfrequenz: 16MHz
;*	Eingaenge:
;*	Ausgaenge:
;*
;*	Informationen zur Funktionsweise:
;*			
;*      Copyright (c) 2012   Guy WEILER         weigu[at]weigu[dot]lu
;*
;*      Hiermit wird unentgeltlich, jeder Person, die eine Kopie der Software
;*      und der zugehoerigen Dokumentationen (die "Software") erhaelt, die
;*      Erlaubnis erteilt, uneingeschraenkt zu benutzen, inklusive und ohne
;*      Ausnahme, dem Recht, sie zu verwenden, kopieren, aendern, fusionieren,
;*      verlegen, verbreiten, unterlizenzieren und/oder zu verkaufen, und
;*      Personen, die diese Software erhalten, diese Rechte zu geben, unter den
;*      folgenden Bedingungen:
;*
;*      ;Der obige Urheberrechtsvermerk und dieser Erlaubnisvermerk sind in alle
;*      ;Kopien oder Teilkopien der Software beizulegen.
;*      ;
;*      ;DIE SOFTWARE WIRD OHNE JEDE AUSDRUECKLICHE ODER IMPLIZIERTE GARANTIE
;*      ;BEREITGESTELLT, EINSCHLIESSLICH DER GARANTIE ZUR BENUTZUNG FUER DEN
;*      ;VORGESEHENEN ODER EINEM BESTIMMTEN ZWECK SOWIE JEGLICHER RECHTS-
;*      ;VERLETZUNG, JEDOCH NICHT DARAUF BESCHRAENKT. IN KEINEM FALL SIND DIE
;*      ;AUTOREN ODER COPYRIGHTINHABER FUER JEGLICHEN SCHADEN ODER SONSTIGE
;*      ;ANSPRUECHE HAFTBAR ZU MACHEN, OB INFOLGE DER ERFUELLUNG EINES
;*      ;VERTRAGES, EINES DELIKTES ODER ANDERS IM ZUSAMMENHANG MIT DER SOFTWARE
;*      ;ODER SONSTIGER VERWENDUNG DER SOFTWARE ENTSTANDEN.
;*
;*******************************************************************************


;-------------------------------------------------------------------------------
;	Einbinden der controllerspezifischen Definitionsdatei
;-------------------------------------------------------------------------------
.NOLIST				;List-Output ausschalten
.INCLUDE "m32u4def.inc"		;AVR-Definitionsdatei einbinden
.LIST				;List-Output wieder einschalten

;-------------------------------------------------------------------------------
;       Organisation des Datenspeichers (SRAM)
;-------------------------------------------------------------------------------
;.DSEG                          ;was ab hier folgt kommt in den SRAM-Speicher
;TAB1:  .BYTE  100              ;100 Byte grosse Tabelle im Datensegment

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;       Programmspeicher (FLASH)     Programmstart nach RESET ab Adr. 0x0000
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.CSEG                           ;was ab hier folgt kommt in den FLASH-Speicher
.ORG	0x0000			;Programm beginnt an der FLASH-Adresse 0x0000
RESET1:	rjmp	INIT		;springe nach INIT (ueberspringe ISR Vektoren)

;-------------------------------------------------------------------------------
;	Sprungadressen fuer die Interrupts organisieren (ISR VECTORS)	
;-------------------------------------------------------------------------------
        ;Vektortabelle (im Flash-Speicher)
.ORG	OC1Aaddr		;Timer/Counter1 Compare Match A
	rjmp	ISRTC1          ;springe zur ISR Timer1
.ORG	URXC1addr		;USART0, Rx Complete
	rjmp	ISRRX           ;springe zur ISR USART0 Empfang
.ORG	UDRE1addr		;USART0 Data Register Empty
	rjmp	ISRTX           ;springe zur ISR USART0 Senden

;-------------------------------------------------------------------------------
;	Initialisierungen und eigene Definitionen
;-------------------------------------------------------------------------------
.ORG    INT_VECTORS_SIZE        ;Platz fuer ISR Vektoren lassen
INIT:
.DEF	Zero = r15		;Register 1 wird zum Rechnen benoetigt
	clr	r15		;und mit Null belegt
.DEF	Tmp1 = r16		;Register 16 dient als erster Zwischenspeicher
.DEF	Tmp2 = r17		;Register 17 dient als zweiter Zwischenspeicher
.DEF	Cnt1 = r18		;Register 18 dient als Zaehler
.DEF	Cnt2 = r19		;Register 19 dient als Zaehler
.DEF	T1Flag = r20		;Register 20 dient als TimerFlag
.DEF	MinFlag = r21		;Register 21 dient als Flag fuer Minutenausgabe
.DEF    WL = r24                ;Register 24 und 25 dienen als universelles
.DEF    WH = r25                ;Doppelregister W und zur Parameteruebergabe

	;Stapel initialisieren (fuer Unterprogramme bzw. Interrupts)
	ldi	Tmp1,LOW(RAMEND)	;RAMEND (SRAM) ist in der Definitions-
	out	SPL,Tmp1		;datei festgelegt
	ldi	Tmp1,HIGH(RAMEND)
	out	SPH,Tmp1
	
	;Flags loeschen
	clr T1Flag
	clr MinFlag
	
	;ADC initialisieren
        ldi     Tmp1,0b11000100 ;REFS = 11 intern 2,56V
        sts     ADMUX,Tmp1      ;ADLAR = 0 rechtsbuendig
                                ;MUX = 00100 unsymetrisch PF4 (ADC4)
        ldi     Tmp1,0b10000111 ;ADEN = 1 (ADC einschalten)
        sts     ADCSRA,Tmp1     ;ADPS = 111 (16MHz/128 = 125kHz)
                                ;andere Bits per default = 0

	;Timer 1 initialisieren
	;Interrupt ein mal pro Sekunde
	;Vorteiler 1024, Zaehlschritte 15625-1 ergeben 1 Sekunden
	clr	Tmp1
	sts	TCCR1A,Tmp1
	ldi	Tmp1,0b00001101	;WGM12 = 1 Teiler = 1024	
	sts	TCCR1B,Tmp1
	ldi	Tmp1,HIGH(15624) ;Achtung High zuerst!	
	sts	OCR1AH,Tmp1
	ldi	Tmp1,LOW(15624)
	sts	OCR1AL,Tmp1
	ldi	Tmp1,0x02	;OCI1A Interrupt erlauben
	sts	TIMSK1,Tmp1	
                                
        rcall	E232INI
	rcall	LCDINI
	rcall	LCDCLR
	
	sei
	
;-------------------------------------------------------------------------------
;	Hauptprogramm
;-------------------------------------------------------------------------------
MAIN:	clr	r6		;r8:r7:r6 fuer Spannung
	clr	r7
	clr	r8
	clr	r10		;r14:r13:r12 fuer Strom
	clr	r11
	clr	r12
	
	;Spannungsmessung
	ldi     Tmp1,0b11000100 ;MUX = 00100 unsymetrisch PF4 (ADC4)
        sts     ADMUX,Tmp1      
        ;fuenf erste Wandlungen verwerfen    
        ldi	Cnt2,5
VLOOP1:	lds	Tmp1,ADCSRA	;eine AD-Wandlung ausloesen
	ori	Tmp1,0x40	;ADSC = 1
	sts	ADCSRA,Tmp1
VLOOP2:	lds	Tmp1,ADCSRA	;Polling bis Wandlung fertig
	sbrc	Tmp1,ADSC
	rjmp	VLOOP2
	lds     Tmp1,ADCL       
        lds     Tmp2,ADCH 
        dec	Cnt2
        brne	VLOOP1
        ;256 Wandlungen    
	ldi	Cnt1,255	
VLOOP3:	lds	Tmp1,ADCSRA	;eine AD-Wandlung ausloesen
	ori	Tmp1,0x40	;ADSC = 1
	sts	ADCSRA,Tmp1
VLOOP4:	lds	Tmp1,ADCSRA	;Polling bis Wandlung fertig
	sbrc	Tmp1,ADSC
	rjmp	VLOOP4
	lds     Tmp1,ADCL       
        lds     Tmp2,ADCH 
        add	r6,Tmp1
        adc	r7,Tmp2
        adc	r8,Zero
        dec	Cnt1
        brne	VLOOP3       
        lsr	r8		; /256
        ror	r7
        ror	r6
        lsr	r8
        ror	r7
        ror	r6
        lsr	r8
        ror	r7
        ror	r6
        lsr	r8
        ror	r7
        ror	r6
        lsr	r8
        ror	r7
        ror	r6
        lsr	r8
        ror	r7
        ror	r6
        lsr	r8
        ror	r7
        ror	r6
        lsr	r8
        ror	r7
        ror	r6

        ;Strommessung
	ldi     Tmp1,0b11001011 ;Differenz ADC1:ADC0
        sts     ADMUX,Tmp1      
	;fuenf erste Wandlungen verwerfen    
        ldi	Cnt2,5
ALOOP1:	lds	Tmp1,ADCSRA	;eine AD-Wandlung ausloesen
	ori	Tmp1,0x40	;ADSC = 1
	sts	ADCSRA,Tmp1
ALOOP2:	lds	Tmp1,ADCSRA	;Polling bis Wandlung fertig
	sbrc	Tmp1,ADSC
	rjmp	ALOOP2
	lds     Tmp1,ADCL       
        lds     Tmp2,ADCH         
        dec	Cnt2
        brne	ALOOP1
        ;256 Wandlungen    
	ldi	Cnt1,255	
ALOOP3:	lds	Tmp1,ADCSRA	;eine AD-Wandlung ausloesen
	ori	Tmp1,0x40	;ADSC = 1
	sts	ADCSRA,Tmp1
ALOOP4:	lds	Tmp1,ADCSRA	;Polling bis Wandlung fertig
	sbrc	Tmp1,ADSC
	rjmp	ALOOP4
	lds     Tmp1,ADCL       
        lds     Tmp2,ADCH 
        add	r10,Tmp1
        adc	r11,Tmp2
        adc	r12,Zero
        dec	Cnt1
        brne	ALOOP3       

        lsr	r12		; /256
        ror	r11
        ror	r10
        lsr	r12
        ror	r11
        ror	r10
        lsr	r12
        ror	r11
        ror	r10
        lsr	r12
        ror	r11
        ror	r10
        lsr	r12
        ror	r11
        ror	r10
        lsr	r12
        ror	r11
        ror	r10
        lsr	r12
        ror	r11
        ror	r10
        lsr	r12
        ror	r11
        ror	r10       

 	;Spannung berechnen und ausgeben
 	;U = (ADC*Vref/1023)*2	Verdopplung, da Spannungsteiler 1:2
 	ldi	Tmp1,LOW(5060)	;Vref gemessen 2,53V (Wert in mV)
        mov	r8,Tmp1		;2,53*2*1000
        ldi	Tmp1,HIGH(5060)	;
        mov	r9,Tmp1
        rcall	MUL22		;Resultat in r5:r4:r3:r2
        mov	r0,r2
        mov	r1,r3
        mov	r2,r4
        mov	r3,r5
        ldi	Tmp1,LOW(1023)
        mov	r4,Tmp1
        ldi	Tmp1,HIGH(1023)
        mov	r5,Tmp1
        rcall	DIV42
        mov	r13,r0		;retten fuer Leistung
        mov	r14,r1
        mov	r24,r0
        mov	r25,r1
        rcall   DEC2A5	
        ldi     r24,PT1+0	;1 Zeile 2 Zeichen
        rcall   LCDSCO          
	lds	r24,TD2A5+1	
	rcall	E232SND
        rcall   LCDSDA        
        ldi     r24,PT1+1	
        rcall   LCDSCO          
	ldi	r24,','
	rcall	E232SND
	rcall   LCDSDA        
        ldi     r24,PT1+2	
        rcall   LCDSCO          
	lds	r24,TD2A5+2	
	rcall	E232SND
        rcall   LCDSDA        
        ldi     r24,PT1+3	
        rcall   LCDSCO          
	lds	r24,TD2A5+3	
	rcall	E232SND
        rcall   LCDSDA        
	lds	r24,TD2A5+4	
	rcall	E232SND
        ldi     r24,PT1+4	
        rcall   LCDSCO          
	ldi	r24,'V'
	rcall	E232SND	
        rcall   LCDSDA        
        ldi     r24,PT1+5	
        rcall   LCDSCO          
	ldi	r24,' '
	rcall	E232SND	
        rcall   LCDSDA         
        ldi     r24,PT1+6	
        rcall   LCDSCO 
        ldi	r24,' '
	rcall   LCDSDA         

	;Strom berechnen und ausgeben
	;I = (ADC*Vref/(200*512))/0,02	Datasheet P306
	;(nach Datenblatt nur 7 Bit sinnvoll)
        mov	r6,r10
	mov	r7,r11
	ldi	Tmp1,6
	sub	r6,Tmp1		;Offsetkorrektur
	sbc	r7,Zero
        ldi	Tmp1,LOW(6325)	;Vref gemessen 2,53V 
        mov	r8,Tmp1		;(2,53/200/0,02)*10000    Wert in mA/10
        ldi	Tmp1,HIGH(6325)
        mov	r9,Tmp1
        rcall	MUL22		;Resultat in r5:r4:r3:r2
        mov	r0,r2
        mov	r1,r3
        mov	r2,r4
        mov	r3,r5
        ldi	Tmp1,LOW(512)
        mov	r4,Tmp1
        ldi	Tmp1,HIGH(512)
        mov	r5,Tmp1
        rcall	DIV42
        mov	r6,r0		;retten fuer Leistung
        mov	r7,r1
        mov	r24,r0
        mov	r25,r1
        rcall   DEC2A5	        
        ldi     r24,PT2+0	
        rcall   LCDSCO          
       	lds	r24,TD2A5+0	
	rcall	E232SND
        rcall   LCDSDA        
        ldi     r24,PT2+1	
        rcall   LCDSCO          
	ldi	r24,','
	rcall	E232SND
	rcall   LCDSDA        
	ldi     r24,PT2+2	
        rcall   LCDSCO          
       	lds	r24,TD2A5+1	
	rcall	E232SND
	rcall   LCDSDA        
	ldi     r24,PT2+3	
        rcall   LCDSCO          
	lds	r24,TD2A5+2	
	rcall	E232SND
	rcall   LCDSDA        
	ldi     r24,PT2+4	
        rcall   LCDSCO          
	lds	r24,TD2A5+3	
	rcall	E232SND
	rcall   LCDSDA        
	lds	r24,TD2A5+4	
	rcall	E232SND	
	ldi     r24,PT2+5	
        rcall   LCDSCO          
	ldi	r24,'A'
	rcall	E232SND	
	rcall   LCDSDA        
	ldi	r24,' '
	rcall	E232SND	

	;Leistung berechnen und ausgeben
	mov	r8,r13
	mov	r9,r14
	rcall	MUL22		;Resultat in r5:r4:r3:r2
        mov	r0,r2
        mov	r1,r3
        mov	r2,r4
        mov	r3,r5
        ldi	Tmp1,LOW(1000)
        mov	r4,Tmp1
        ldi	Tmp1,HIGH(1000)
        mov	r5,Tmp1
        rcall	DIV42
	mov	r24,r0
        mov	r25,r1
        rcall   DEC2A5	
        ldi     r24,PT3+0	
        rcall   LCDSCO          
      	lds	r24,TD2A5+0	
      	rcall	E232SND
	rcall   LCDSDA        
        ldi     r24,PT3+1	
        rcall   LCDSCO          
	ldi	r24,','
	rcall	E232SND
	rcall   LCDSDA        
        ldi     r24,PT3+2	
        rcall   LCDSCO          
       	lds	r24,TD2A5+1	
	rcall	E232SND
	rcall   LCDSDA        
        ldi     r24,PT3+3	
        rcall   LCDSCO          
	lds	r24,TD2A5+2	
	rcall	E232SND
	rcall   LCDSDA        
        ldi     r24,PT3+4	
        rcall   LCDSCO          
	lds	r24,TD2A5+3	
	rcall	E232SND
	rcall   LCDSDA        
	lds	r24,TD2A5+4	
	rcall	E232SND	
        ldi     r24,PT3+5	
        rcall   LCDSCO          
	ldi	r24,'W'
	rcall	E232SND
        rcall   LCDSDA        

	ldi	r24,0x0D
	rcall	E232SND
	
	ldi	Cnt2,60		;Minutenzaehler (60s)
	    
TLOOP:	cpi	T1Flag,1
	brne	TLOOP
	;Befehle abarbeiten (Flags ueberpruefen)
	;"?" Help
COMM1:	lds	Tmp1,RXFLAG
        cpi	Tmp1,0x1F	;? Hilfe angefordert?
        brne	COMM2		;"command list" ausgeben  
        rcall	E232CL		;       
        sts	RXFLAG,Zero	;RXFLAG zuruecksetzen

        ;"s" Timer loest 1* pro Sekunde aus versenden
COMM2:  lds	Tmp1,RXFLAG
	cpi	Tmp1,0x2F	;
        brne	COMM3
	clr	MinFLAG
        sts	RXFLAG,Zero	;RXFLAG zuruecksetzen

        ;"m" Timer loest 1* pro Minute
COMM3:  lds	Tmp1,RXFLAG
	cpi	Tmp1,0x3F	;
        brne	COMM4
	ldi	MinFLAG,1
        sts	RXFLAG,Zero	;RXFLAG zuruecksetzen

COMM4:  cpi	MinFlag,1
	brne	BACK
	dec	Cnt2
	breq	BACK
	clr	T1Flag
	rjmp	TLOOP
	
BACK:	clr	T1Flag
        rjmp	MAIN	;Endlosschleife

;-------------------------------------------------------------------------------
;	Unterprogramme und Interrupt-Behandlungsroutinen
;-------------------------------------------------------------------------------
;Timer 1 ISR Interrupt jede Sekunde oder Minute
;Setz Flag
ISRTC1: push    r16             ;rette r16 (als lokale Variable verwendet)
        in      r16,SREG        ;SREG-Register (Flags) retten
        push    r16             

	ldi	T1Flag,1
	
	pop     r16             ;SREG-Register (Flags) wiederherstellen
        out     SREG,r16        ;
        pop     r16             ;r16 wiederherstellen
        reti    	

.INCLUDE "lib/SR_TIME_16M.asm"		;Zeitschleifenbibliothek einbinden
.INCLUDE "lib/SR_EIA232.asm"     	;EIA 232-Bibliothek einbinden
.INCLUDE "lib/SR_LCD_2x16_4Bit.asm"     ;LCD-Bibliothek einbinden
.INCLUDE "lib/SR_DIV_MUL.asm"     	;Rechen-Bibliothek einbinden
.INCLUDE "lib/SR_RECODE.asm"     	;Kodier-Bibliothek einbinden

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.EXIT				;Ende des Quelltextes
